/* 

Code by: RoAnWo
Date: 08.06.2025

This Arduino sketch controls three stepper motors and an LED strip based on the direction of detected sounds.
Three sound sensors are arranged in a triangle; the code estimates the direction of a sharp sound and moves the motors accordingly.
The LED strip provides visual feedback for each detected sound event.

Feel free to use or modify the code.

Reference:
Clyde Lettsome`s Sound Location Finder: https://clydelettsome.com/blog/2019/09/08/my-weekend-project-sound-location-finder/
- this "Sound Location Finder" code uses 2 microphones and works a bit differtently, but was used as a starting point for this Project.

GPT was used for Bugfixing, added comments and cleanup of Sketch

*/



#include <math.h>

// Sound sensor pins
const int soundPins[3] = {A0, A1, A2};
unsigned long triggerTimes[3] = {0, 0, 0};
bool triggered[3] = {false, false, false};
unsigned long detectionWindow = 100000;  // 100ms window
unsigned long firstTriggerTime = 0;
byte firstSensor = 255;
const int ledStripPin = 13;     // New MOSFET-controlled LED strip

// CNC Shield V3 stepper pins
const int stepPins[3] = {2, 3, 4}; 
const int dirPins[3] = {5, 6, 7};
const int enPin = 8;
const int stepsPerRevolution = 200;

float interpFactor = .7; // 1.0 = linear, >1 = sharper, <1 = softer
bool useDefaultTransition = true; // Toggle between direct movement between sound sources and through hanging position (if it has a frame where it cant move 360 degrees) 

// track current steps and las sound event for moving from one position to another/ returning to "default" (hanging) state
int currentSteps[3] = {0, 0, 0};
unsigned long lastSoundTime = 0;
unsigned long lastMoveTime = 0; // Tracks last motor movement

// System parameters
const float OFFSET = 150.0; 
const float SENSOR_DISTANCE = 0.25; // 25cm between sensors
float moveAmount = 5.0; // 1.0 = full rotation per weight 1.0, 0.5 = half rotation, etc.
void calculateMotorWeights(float theta, float weights[3]) {
  // Normalize angle to [0, 360)
  theta = fmod(theta + 360.0, 360.0);

  weights[0] = weights[1] = weights[2] = 0.0;

  if ((theta >= 0 && theta < 60) || (theta >= 300 && theta < 360)) {
    weights[0] = 1.0;
    if (theta >= 0 && theta < 60) {
      weights[1] = pow(theta / 60.0, interpFactor);
    } else {
      weights[2] = pow((theta - 300) / 60.0, interpFactor);
    }
  } else if (theta >= 60 && theta < 120) {
    weights[1] = 1.0;
    weights[0] = pow((120 - theta) / 60.0, interpFactor);
  } else if (theta >= 120 && theta < 180) {
    weights[1] = 1.0;
    weights[2] = pow((theta - 120) / 60.0, interpFactor);
  } else if (theta >= 180 && theta < 240) {
    weights[2] = 1.0;
    weights[1] = pow((240 - theta) / 60.0, interpFactor);
  } else if (theta >= 240 && theta < 300) {
    weights[2] = 1.0;
    weights[0] = pow((theta - 240) / 60.0, interpFactor);
  }

  // Clamp to [0,1]
  for(int i=0; i<3; i++) {
    weights[i] = constrain(weights[i], 0.0, 1.0);
  }
}

void moveMotors(int deltaSteps[3]) {
  digitalWrite(enPin, LOW);
  int absSteps[3];
  int maxSteps = 0;

  // Set directions and calculate absolute steps
  for(int i=0; i<3; i++) {
    if(deltaSteps[i] > 0) {
      digitalWrite(dirPins[i], LOW); // Forward
    } else {
      digitalWrite(dirPins[i], HIGH); // Backward
    }
    absSteps[i] = abs(deltaSteps[i]);
    if(absSteps[i] > maxSteps) maxSteps = absSteps[i];
  }

  // Move motors
  for(int s=0; s<maxSteps; s++) {
    for(int i=0; i<3; i++) {
      if(s < absSteps[i]) {
        digitalWrite(stepPins[i], HIGH);
      }
    }
    delayMicroseconds(800);
    for(int i=0; i<3; i++) digitalWrite(stepPins[i], LOW);
    delayMicroseconds(800);
  }


  digitalWrite(enPin, HIGH);
  lastMoveTime = millis();
}


void setup() {
  Serial.begin(115200);
  for(int i=0; i<3; i++) {
    pinMode(soundPins[i], INPUT_PULLUP);
    pinMode(stepPins[i], OUTPUT);
    pinMode(dirPins[i], OUTPUT);
  }
  pinMode(enPin, OUTPUT);
  pinMode(ledStripPin, OUTPUT);  // Initialize MOSFET control pin
  digitalWrite(enPin, LOW);
  Serial.println("System initialized");
}

void estimateDirection(long delays[3]) {
  float t[3];
  for (byte i = 0; i < 3; i++) {
    t[i] = triggered[i] ? (delays[i] * 1e-6) : NAN;
  }

  float dt01 = t[0] - t[1];
  float dt02 = t[0] - t[2];
  
  float dx = (dt01 * SENSOR_DISTANCE)/2.0;
  float dy = (dt02 * SENSOR_DISTANCE * sqrt(3))/3.0;

  float raw_angle = atan2(dy, dx) * 180.0 / M_PI;
  float angle = fmod(raw_angle + OFFSET + 360.0, 360.0);

  Serial.print("Direction: ");
  Serial.print(angle, 1);
  Serial.println("° (0°=motor 0, 120°=motor 1, 240°=motor 2)");

  float weights[3];
  calculateMotorWeights(angle, weights);

  Serial.print("Motor weights: ");
  for (int i = 0; i < 3; i++) {
    Serial.print(weights[i], 2);
    Serial.print(i < 2 ? ", " : "\n");
  }

  // Calculate target steps for each motor
  int targetSteps[3];
  for(int i=0; i<3; i++) {
    targetSteps[i] = round(weights[i] * stepsPerRevolution * moveAmount);
  }

  // Movement logic based on transition mode
  if(useDefaultTransition) {
    // Two-stage movement: default position -> target
    int returnDelta[3];
    for(int i=0; i<3; i++) {
      returnDelta[i] = -currentSteps[i]; // Return to zero
    }
    moveMotors(returnDelta);
    
    // Update position tracking to zero
    for(int i=0; i<3; i++) currentSteps[i] = 0;
    
    // Move from zero to target
    int targetDelta[3];
    for(int i=0; i<3; i++) {
      targetDelta[i] = targetSteps[i];
    }
    moveMotors(targetDelta);
  } else {
    // Direct movement: current position -> target
    int deltaSteps[3];
    for(int i=0; i<3; i++) {
      deltaSteps[i] = targetSteps[i] - currentSteps[i];
    }
    moveMotors(deltaSteps);
  }

  // Update current position (same for both modes)
  for(int i=0; i<3; i++) {
    currentSteps[i] = targetSteps[i];
  }
}


void loop() {
  for(byte i=0; i<3; i++) {
    if(digitalRead(soundPins[i]) == LOW && !triggered[i]) {
      triggerTimes[i] = micros();
      triggered[i] = true;
      if(firstSensor == 255) {
        firstSensor = i;
        firstTriggerTime = triggerTimes[i];
      }
    }
  }

  if(firstSensor != 255 && (micros() - firstTriggerTime) > detectionWindow) {
    Serial.println("\n=== Sound Event ===");

    long delays[3];
    byte activeSensors = 0;
    for(byte i=0; i<3; i++) {
      delays[i] = triggered[i] ? (triggerTimes[i] - firstTriggerTime) : 0;
      Serial.print("Sensor ");
      Serial.print(i);
      Serial.println(triggered[i] ? " ✔" : " ❌");
      if(triggered[i]) activeSensors++;
    }

    if(activeSensors > 2) {
      estimateDirection(delays);
    } else {
      Serial.println("Ignored: Need 3 sensors");
    }

    // Reset
    firstSensor = 255;
    memset(triggered, 0, sizeof(triggered));
    memset(triggerTimes, 0, sizeof(triggerTimes));

    // Activate LEDs
    digitalWrite(ledStripPin, HIGH);
    delay(1000);
    digitalWrite(ledStripPin, LOW);

    // Reset timeout timer after every sound event (optional, keep for legacy)
    lastSoundTime = millis();
  }

  // Print time since last movement every second (for debugging)
  static unsigned long lastPrintTime = 0;
  if (millis() - lastPrintTime > 1000) {
    lastPrintTime = millis();
    Serial.print("Time since last move: ");
    Serial.print((millis() - lastMoveTime) / 1000.0, 1);
    Serial.println(" seconds");
  }

  // Timeout check: move to default if no movement for X seconds
  // (now based on lastMoveTime, not lastSoundTime)
  if ((currentSteps[0] != 0 || currentSteps[1] != 0 || currentSteps[2] != 0) &&
      (millis() - lastMoveTime >= 10000)) { //ADJUST THIS FOR after how long of no event to move back into default position
    Serial.println("\nNo movement for 10s. Returning to initial position.");
    int deltaSteps[3];
    for(int i=0; i<3; i++) {
      deltaSteps[i] = -currentSteps[i];
    }
    moveMotors(deltaSteps);
    for(int i=0; i<3; i++) {
      currentSteps[i] = 0;
    }
    // Update last movement time to prevent immediate retrigger
    lastMoveTime = millis();
  }
}
